#!/usr/bin/python

import argparse
import re
import os
import sys
from decimal import Decimal
import json
import shutil

import zipfile
import tarfile
from io import BytesIO

from collections import defaultdict


def find_matching_domain(prob_path):
    d = os.path.dirname(prob_path)
    p = os.path.basename(prob_path)[:-5]
    res = None
    for x in ["domain.pddl", p + "-domain.pddl", "domain-" + p + ".pddl", p.replace("problem", "domain") + ".pddl"]:
        res = os.path.join(d, x)
        if os.path.exists(res):
            break
        res = None
    if res is None:
        sys.stderr.write("Could not find domain file associated with %s\n" % prob_path)
    assert(res != None)
    return res


def apply_horizon_changes_to_domain(domain_path, costs):
    res = []
    with open(domain_path) as f:
        content = f.read()

    for cost in re.finditer(r"\(increase \(total-cost\) (\d+)\)", content):
        costs.append(int(cost.group(1)))

    costs = list(sorted(set(costs)))

    i = 0
    j = content.find("(:action")
    res.append(\
        re.sub("\(:types", "(:types budget-value - object", \
            re.sub("\(:predicates", "(:predicates (budget ?h - budget-value) (budget-decrement ?h0 ?h1 ?h2 - budget-value)", \
                re.sub("\(:constants", "(:constants %s - budget-value " % " ".join("bud%d" % c for c in costs), \
                    content[:j]))))

    if not ":types" in res[0]:
        k = res[0].find("(:constants")
        if k < 0:
            k = res[0].find("(:predicates")
        res[0] = res[0][:k] + " (:types budget-value - object) " + res[0][k:]

    if not ":constants" in res[0]:
        k = res[0].find("(:predicates")
        res[0] = res[0][:k] + " (:constants %s - budget-value) " % " ".join("bud%d" % c for c in costs) + res[0][k:]

    while j >= 0:
        i = j
        j = content.find("(:action", i + 8)
        action = content[i:] if j < 0 else content[i:j]
        cost = re.search(r"\(increase \(total-cost\) (\d+)\)", action)
        if cost != None:
            if not ":parameters" in action:
                k = action.find(":precondition")
                action = action[:k] + ":parameters (?hcur ?hnew - budget-value)\n" + action[k:]
            else:
                action = re.sub(":parameters\s*\((.*)\)", r":parameters (?hcur ?hnew - budget-value \1)", action)
            if re.search(":precondition\s*\(and", action) == None:
                if not ":precondition" in action:
                    print(action)
                    assert(False)
                action = re.sub(":precondition\s*\((.*)\)", r":precondition (and (budget ?hcur) (budget-decrement ?hcur bud%s ?hnew) (\1))" % cost.group(1), action)
            else:
                action = re.sub(":precondition\s*\(and", ":precondition (and (budget ?hcur) (budget-decrement ?hcur bud%s ?hnew)" % cost.group(1), action)

            assert(re.search(":effect\s*\(and", action) != None)
            action = re.sub(":effect\s*\(and", ":effect (and (not (budget ?hcur)) (budget ?hnew)", action)
        res.append(action)
    return "".join(res)


def apply_horizon_changes_to_problem(problem_path, horizon, costs):
    with open(problem_path) as f:
        content = f.read()
    costs = sorted(set(costs))
    obs = " ".join(["bud%d" % i for i in range(horizon + 1) if i not in costs])
    init = " ".join(["(budget bud%d)" % horizon] + ["(budget-decrement bud%d bud%d bud%d)" % (i+c, c, i) for c in costs for i in range(horizon - c + 1)])
    if not "(:objects" in content:
        i = content.find("(:init")
        res = []
        res.append(content[:i])
        res.append("\n")
        res.append("(:objects %s - budget-value)\n" % obs)
        res.append("(:init %s" % init)
        res.append(content[i+6:])
        return "".join(res)
    else:
        return re.sub("\(:objects", "(:objects %s - budget-value" % obs, \
                re.sub("\(:init", "(:init %s" % init, content))

        
def replace_default_horizon(problem_path, horizon):
    with open(problem_path) as f:
        content = f.read()

    res = []

    objs_i = content.find("(:objects")
    init_i = content.find("(:init")
    goal_i = content.find("(:goal")
    assert(objs_i >= 0 and init_i >= 0 and goal_i >= 0)
    assert(objs_i < init_i and init_i < goal_i)

    res.append(content[:objs_i])

    regx = re.compile("(bud\d+\s*)+ - budget-value")
    assert(regx.search(content[objs_i:init_i]) != None)
    res.append(regx.sub(" ".join(["bud%d" % i for i in range(horizon + 1)]) + " - budget-value", content[objs_i:init_i]))

    regx = re.compile(r"\s*\(budget-decrement bud\d+ bud\d+\)\s*")
    assert(regx.search(content[init_i:goal_i]) != None)
    init = regx.sub("", content[init_i:goal_i])
    
    regx = re.compile(r"\(budget bud\d+\)")
    assert(regx.search(init) != None)
    init = regx.sub("\n".join(["(budget bud%d)" % horizon] + ["(budget-decrement bud%d bud%d)" % (i+1, i) for i in range(horizon)]), init)

    res.append(init)
    res.append(content[goal_i:])

    return "".join(res)


class CollectionWriter(object):
    def add(self, domain, f, content):
        raise NotImplementedError
    def open(self):
        pass
    def close(self):
        pass

class DirectoryWriter(CollectionWriter):
    def __init__(self, base_path):
        self.base_path = base_path
    def add(self, domain, f, content):
        if not os.path.isdir(os.path.join(self.base_path, domain)):
            os.makedirs(os.path.join(self.base_path, domain))
        with open(os.path.join(self.base_path, domain, f), "w") as fh:
            fh.write(content)

class TarWriter(CollectionWriter):
    def __init__(self, path, ext):
        self.path = path
        self.ext = ext
    def add(self, domain, f, content):
        data = content.encode('ascii')
        info = tarfile.TarInfo(name=os.path.join(domain, f))
        info.size = len(data)
        self.fh.addfile(info, BytesIO(data))
    def open(self):
        self.fh = tarfile.open(self.path, "w:" + self.ext)
    def close(self):
        self.fh.close()
        self.fh = None


def from_to(info, scales, src, to):
    NOT_FOUND = -10
    UNSATISFIABLE = -20
    def get_min_horizon_or_error_code(domain, problem):
        if not domain in info \
                or not problem in info[domain]:
            return NOT_FOUND
        dat = info[domain][problem]
        if "unsolvable" in dat and dat["unsolvable"]:
            return UNSATISFIABLE
        if "minimal_budget" in dat:
            return dat["minimal_budget"]
        return NOT_FOUND


    for domain in os.listdir(src):
        if not os.path.isdir(os.path.join(src, domain)):
            continue
        status = {}

        path = os.path.join(src, domain)
        has_single_domain = sum(["domain" in x for x in os.listdir(path)]) == 1

        for scale in scales:
            gamma = Decimal(scale)
            affix = str((gamma * 100).to_integral())
            idad = domain + "-" + affix

            if not os.path.exists(os.path.join(to, idad)):
                os.makedirs(os.path.join(to, idad))
            
            costs = []

            if has_single_domain:
                assert(os.path.exists(os.path.join(path, "domain.pddl")))
                with open(os.path.join(to, idad, "domain.pddl"), "w") as f:
                    f.write(apply_horizon_changes_to_domain(os.path.join(path, "domain.pddl"), costs))

            for prob in os.listdir(path):
                probx = os.path.basename(prob)
                if not "domain" in prob:
                    minh = get_min_horizon_or_error_code(domain, probx)
                    status[probx] = minh
                    if minh >= 0:
                        if not has_single_domain:
                            costs = []
                            dom = find_matching_domain(os.path.join(path, probx))
                            with open(os.path.join(to, idad, os.path.basename(dom)), "w") as f:
                                f.write(apply_horizon_changes_to_domain(dom, costs))
                        with open(os.path.join(to, idad, os.path.basename(prob)), "w") as f:
                            f.write(apply_horizon_changes_to_problem(os.path.join(path, prob), int((gamma * minh).to_integral()), costs))

        for x in sorted(status.keys()):
            if status[x] < 0:
                reason = "minimal budget value was not found" if status[x] == NOT_FOUND else (\
                        "task is unsolvable" if status[x] == UNSATISFIABLE else (\
                        "due to some unknown reason"))
                sys.stderr.write("Skipping %s:%s (%s)\n" % (domain, x, reason))


def main():
    p = argparse.ArgumentParser()
    p.add_argument("database")
    p.add_argument("source", help="Where the benchmark collection should be stored. Can be either folder or archive name.")
    p.add_argument("destination", help="Where the benchmark collection should be stored. Can be either folder or archive name.")
    p.add_argument("--constrainedness", "-c", default="1.0,1.2,1.4,1.6,1.8,2.0")
    args = p.parse_args()
    with open(args.database) as f:
        info = json.loads(f.read())
    from_to(info, args.constrainedness.split(','), args.source, args.destination)



if __name__ == "__main__":
    main()

